arXiv is now an independent nonprofit! Learn more
License: CC BY 4.0
arXiv:2609.14845v1 [cs.AR] 13 Sep 2026

Accurate Models of AMD Matrix Cores

Faizan A. Khattak email: f.a.khattak@leeds.ac.uk Affiliation: University of Leeds, Leeds, Leeds, UK , Mantas Mikaitis email: mmikaitis@leeds.ac.uk Affiliation: University of Leeds, Leeds, Leeds, UK and Carlo J. Graziani email: cgraziani@anl.gov Affiliation: Argonne National Laboratory, Lemont, IL, USA
© none
Abstract.

Matrix multipliers available on recent GPUs do not conform with the IEEE 754 floating point standard. Features of matrix multipliers differ across vendors and architectures of the same vendor, such as accumulator width, rounding behaviour, normalisation points, intermediate underflow and overflow logic, the handling of subnormals, and the treatment of special inputs. As a result, reproducibility of small matrix multiplier results across devices is not possible and cannot be achieved by software control. Implementation details of matrix multipliers are not documented, making it difficult to interpret discrepancies in the computed results. We characterise the numerical behaviour of matrix multipliers across three AMD GPU architectures: CDNA 1, CDNA 2, and CDNA 3, using the MI100, MI210/250, and MI300A/300X GPUs, respectively. We design test vectors to target numerical features for all supported input formats and provide the derivation and the reasoning for why each vector allows to determine a particular numerical feature based on the outputs of the devices. MATLAB-based software models of the matrix multipliers are then developed for each architecture and validated for bit-level reproducibility against hardware using a randomized test suite consisting of 10 million sets of random input vectors. To achieve this, we applied a previously developed technique to iteratively refine the accuracy of the models in a loop, by randomized testing followed by test-refinement until the model matches the hardware for every test case. Finally, as a proof of concept for what experimental research can be done with the models, we have utilised them in two demonstrative numerical applications, quantifying application-level accuracy differences between AMD matrix cores and the NVIDIA tensor cores.

Keywords: 
Matrix cores, mixed-precision computing, matrix multiply, inner product, IEEE 754 standard arithmetic, OCP low-precision formats

The software associated with this paper, the MATLAB Tensor Core v0.6, which includes all three CDNA architecture matrix core models as well as many NVIDIA tensor core models and a generalised model that can be used to instantiate custom variants of matrix multipliers, is available on GitHub: https://github.com/north-numerical-computing/MATLAB-tensor-core.

1. Introduction

Early work by Hickmann and Bradford (Hickmann and Bradford, 2019) on the NVIDIA V100 GPU showed how to determine several numerical features of matrix multipliers that were not documented, such as the rounding mode and internal precision of the accumulator. Subsequently, (Fasi et al., 2021; Li et al., 2024; Valpey et al., 2025) conducted experiments on the successors to the V100, the A100, T4, and the H100, as well as GPUs from AMD, the MI100, and MI250X, demonstrating differences between the V100, A100 and H100, for example, in the accumulator’s precision within the dot products, by showing that 24-bit, 25-bit and 26-bit accumulation was used, respectively. Due to the absence of documented numerical behaviour, existing software-based matrix multiplier models may fail to accurately represent hardware implementations. In many cases, simulations rely on IEEE-compliant models, which do not faithfully capture the behaviour of these specialised units. Developing accurate software models is therefore essential—not only for understanding hardware behaviour but also for enabling reliable scientific computing and ensuring consistency across diverse platforms.

In our previous work (Khattak and Mikaitis, 2026) we investigated the numerical features of NVIDIA tensor cores. Using a combination of test vectors targeting specific floating-point behaviours and the generalised numerical feature testing (GNFT) framework from (A. Khattak and Mikaitis, 2025), we identified a wide range of numerical features. While most features were extracted using GNFT, some required ensemble-based random testing, revealing new features that require the refinement of the models. This study covered multiple architectures, including Volta, Ampere, Ada Lovelace, Hopper, and Blackwell. Based on these findings, we developed MATLAB-based models that achieve bit-level accuracy and allow both architecture-specific and user-defined configurations.

More recently, (Xie et al., 2025) investigated numerical features of both NVIDIA and AMD GPUs. However, their study reports results for various architectures without providing feature-targeted test vectors, making it difficult to independently verify the reported numerical behaviours and assess the reproducibility of the results.

Our contributions are as follows.

  1. (1)

    We investigate matrix multipliers on AMD GPUs across CDNA 1 (MI100), CDNA 2 (MI210, MI250), and CDNA 3 (MI300A, MI300X) architectures. Where previous literature (Xie et al., 2025) only specified the behaviour, we provide a detailed characterisation of their numerical features, together with test vectors and explanations of their effectiveness. This enables the testing methodology to be extended to future AMD architectures. Our findings also reveal differences from the characterisation reported in (Xie et al., 2025), which we discuss in detail in the paper. The repository accompanying (Xie et al., 2025) on GitHub11 1 https://github.com/microsoft/MMA-Sim, however, is consistent with our findings.

  2. (2)

    We develop MATLAB-based software models for all three CDNA architectures with flexibility to alter various features for numerical experimentation. We expanded the MATLAB Tensor Core software, which previously consisted of more than 10 models of NVIDIA GPU tensor cores (Khattak and Mikaitis, 2026), with CDNA matrix cores.

  3. (3)

    We demonstrate the impact of these differences using multi-word arithmetic for emulating high-precision GEMM via low precision (Mary and Mikaitis, 2025) and broadband signal-processing algorithm (Redif et al., 2015), where GEMMs constitute the dominant computational workload.

2. Notations and Definitions

Table 1 shows characteristics of various floating-point formats available on the latest AMD architecture. Hereafter we refer to floating-point formats with the following short-hand names: fp8 (either of E4M3 or E5M2), fp16 (binary16 of IEEE 754 (IEEE Computer Society, 2019)), bf16 (bfloat16), tf19 (commonly referred to as tensorfloat32), fp32 (binary32 of IEEE 754), and fp64 (binary64 of IEEE 754). As a side note, AMD documents refer to E5M2 and E4M3 by fp8, and bf8, respectively.

Table 1. Floating-point formats which are available as input formats to matrix cores, in the GPU devices based on the AMD CDNA 1–4 architectures. CDNA 4 is included for completeness, but we do not have access to this architecture at the time of writing.
Format precision (ff) min norm. pos. max pos. CDNA Arch.
binary64 (double) 53 2−10222^{-1022} ∼1.798×10308\sim 1.798\times 10^{308} 1/2/3/4
binary32 (single) 24 2−1262^{-126} ∼3.403×1038\sim 3.403\times 10^{38} 1/2/3/4
tf19 (19-bit) 11 2−1262^{-126} ∼3.401×1038\sim 3.401\times 10^{38} 3/4
bfloat16 8 2−1262^{-126} ∼3.389×1038\sim 3.389\times 10^{38} 1/2/3/4
binary16 (half) 11 2−142^{-14} 6550465504 1/2/3/4
fp8-E4M3 (FNUZ/OCP) 4 2−7/2−62^{-7}/2^{-6} 240/448240/448 3/4
fp8-E5M2 (FNUZ/OCP) 3 2−15/2−142^{-15}/2^{-14} 57344/5734457344/57344 3/4

Take two matrices A∈ℝm×kA\in\mathbb{R}^{m\times k}, B∈ℝk×nB\in\mathbb{R}^{k\times n}. The AMD matrix cores accelerate a matrix fused-multiply-add (MFMA) operation

D=A​B+C∈ℝm×n.\displaystyle D=AB+C\in\mathbb{R}^{m\times n}.

In the scope of MFMA, AA, BB are input matrices and CC, DD are accumulator and output matrices, respectively. For simplicity, if we denote with di​jd_{ij} the element at iith row and jjth column of DD, we can express di​jd_{ij} as the inner product between the iith row of AA and jjth column of BB and an addition of a corresponding element in CC as

di​j=∑ℓ=1kai​ℓ​bℓ​j+ci​j.\displaystyle d_{ij}=\sum_{\ell=1}^{k}a_{i\ell}b_{\ell j}+c_{ij}.

To focus on the inner product as an underlying operation, rather than on particular elements of DD, we will omit the subscripts:

(1) d=∑ℓ=1kaℓ​bℓ+c=∑ℓ=1kpℓ+c.\displaystyle d=\sum_{\ell=1}^{k}a_{\ell}b_{\ell}+c=\sum_{\ell=1}^{k}p_{\ell}+c.

The operations in (1) can be implemented either as a sequential fused multiply–accumulate (SFMA) or using a hardware-supported multi-term floating-point addition strategy (Mikaitis, 2024; Khattak and Mikaitis, 2026; Tenca, 2009; Alexandridis and Dimitrakopoulos, 2025). Moreover, multiple implementations of multi-term floating-point addition exist (Khattak and Mikaitis, 2026; Xie et al., 2025). Such implementations may employ additional alignment bits during significand alignment, which we denote by neab{n_{\mathrm{eab}}}. For architectures that accumulate the cc term early, neab{n_{\mathrm{eab}}} directly represents the number of additional alignment bits. In contrast, for architectures that accumulate cc after the product accumulation (late accumulation), this interpretation does not apply directly. In this case, neab{n_{\mathrm{eab}}} refers only to the alignment of the product terms, whereas the alignment of cc with the accumulated product sum, whether normalized or denormalized, is described explicitly for each architecture. Throughout this work, for any many-term dot-product accumulation, we define the number of product terms accumulated in a single multi-term addition, regardless of late or early cc accumulation, as the FMA size (NFMA{N_{\mathrm{FMA}}}) (Khattak and Mikaitis, 2026). Throughout this paper, the term fractional bits is used exclusively to denote the fractional-bit component of a fixed-point representation.

3. Hardware-Guided Numerical Feature Detection and Model Refinement

Numerical feature testing (A. Khattak and Mikaitis, 2025; Li et al., 2024; Fasi et al., 2021; Hickmann and Bradford, 2019) relies on a predefined numerical feature space and then uses targeted feature test vectors to determine whether particular features are present in a given matrix or tensor core architecture. While features may naturally exhibit interdependencies, targeted feature testing alone cannot be considered entirely reliable. Moreover, if a feature present in the hardware is not included in the feature space, it cannot be identified by targeted tests restricted to that space.

To address this limitation, we designed (Khattak and Mikaitis, 2026) an iterative model refinement strategy that combines numerical feature testing with randomized testing against actual hardware. The approach constructs a model from the currently identified features and compares its outputs with those of the hardware over carefully selected randomized test vector sets. A mismatch indicates that the current feature space or model is incomplete, prompting further investigation and the introduction of additional features into the refinement process. This iterative procedure continues until the model agrees with the hardware for the test vectors considered.

For the proposed modelling framework, we construct an initial comprehensive feature space by incorporating all features considered in (A. Khattak and Mikaitis, 2025), together with additional features identified in NVIDIA GPU tensor core architectures through randomized testing in (Khattak and Mikaitis, 2026). We also include all features reported for AMD GPUs in (Xie et al., 2025). These features include, among others, product and accumulation precision, rounding mode, and the number of alignment bits used when aligning product significands. Because the input space of matrix core operations is prohibitively large to explore exhaustively, randomized testing requires an intelligent sampling strategy to efficiently identify discrepancies between the hardware and the model. Such discrepancies provide evidence of potentially missing numerical features and guide subsequent refinement of both the feature space and the model. In Section 5, we discuss how such a finite set of test vectors is constructed.

4. Results

We have analyzed compute DNA (CDNA) 1, 2 and 3 architectures in this paper by running numerical feature detection similar to (A. Khattak and Mikaitis, 2025). We have used MI100, MI210, MI250, MI300, and MI300X AMD data centre GPUs. In all CDNA architectures, matrix cores can be invoked via specific intrinsic MFMA instructions: mfma_In_shape_Out where In can take on fp32, fp16, and bf16 input formats while Out is strictly in fp32, and matrix shape is in format mxnxk which can take on values that vary with input formats.22 2 Examples: ___builtin_amdgcn_mfma_f32_16x16x4f32, ___builtin_amdgcn_mfma_f32_16x16x16f16

4.1. CDNA 1 Architecture

We address each input format case separately.

4.1.1. CDNA1 fp32 input

The supported matrix sizes for the shape argument are 32x32x1 16x16x1, 4x4x1, 32x32x2 and 16x16x4. We adopted the identical numerical feature determining strategy as that of the previous work (Fasi et al., 2021; A. Khattak and Mikaitis, 2025; Li et al., 2024) and we detected that fp32 input matrix cores are SFMAs (see Section 2). Initially, the rounding mode is determined from p1p_{1}, cc, and the resulting dd, as defined in 1. For p1=±1p_{1}=\pm 1 and c=±{2−23+2−24, 2−24}c=\pm\{2^{-23}+2^{-24},\,2^{-24}\}, with all inputs having the same sign, d=±{1+2−22, 1}d=\pm\{1+2^{-22},\,1\}, indicating round-to-nearest, ties-to-even (RNE). To determine whether the operation is an SFMA-like, we first permuted 11, 2−232^{-23}, and 2−242^{-24} among p1p_{1}, p2p_{2}, and cc. We obtained d=1+2−23d=1+2^{-23} when p2=2−23p_{2}=2^{-23}; otherwise, we obtained d=±(1+2−22)d=\pm(1+2^{-22}). These results suggest that cc and p1p_{1} are added first, and their sum is then added to p2p_{2}. We then permuted 11, 2−23+2−242^{-23}+2^{-24}, and 2−23+2−242^{-23}+2^{-24} among p1p_{1}, p2p_{2}, and p3p_{3}. For all permutations except the case p1=p2=2−23+2−24p_{1}=p_{2}=2^{-23}+2^{-24} and p3=1p_{3}=1, we obtained d=1+2−21d=1+2^{-21}. For this exceptional case, we obtained d=1+2−22+2−23d=1+2^{-22}+2^{-23}. These results demonstrate that the fp32 input matrix core internally implements an SFMA for all supported matrix sizes, i.e., (…​(((c+p1)+p2)+p3)+…)(\dots(((c+p_{1})+p_{2})+p_{3})+\dots). Although, in an SFMA, it can be difficult to determine how products are computed and kept, we still perform a series of following tests.

  1. (1)

    By setting p1=±(2−149+2−150)p_{1}=\pm(2^{-149}+2^{-150}) via a1=±(2−126+2−127)a_{1}=\pm(2^{-126}+2^{-127}), b1=2−23b_{1}=2^{-23}, and pi=c=0p_{i}=c=0 for all i≠1i\neq 1, we obtain d=±2−148d=\pm 2^{-148}. This result indicates that either the products are rounded to FP32 using RNE, or the RNE operation is applied only at the end of the SFMA operation.

  2. (2)

    Next, we set p1=2−149p_{1}=2^{-149} and p2=2−149+2−150p_{2}=2^{-149}+2^{-150}, and obtain d=2−148d=2^{-148}. This result confirms that the previous observation was caused by the final rounding step after the SFMA operation. Otherwise, if the products were rounded before the addition, we would have obtained d=2−148+2−149d=2^{-148}+2^{-149}. Therefore, this indicates that the intermediate products are retained in full precision before the addition stage.

  3. (3)

    For further confirmation, we set p1=2−149p_{1}=2^{-149} and p2=2−148+2−150p_{2}=2^{-148}+2^{-150} to prevent any possible upward rounding of the products. We obtain d=2−147d=2^{-147}, which further confirms that the rounding observed in the first case was due to the final rounding step. Hence, the products are preserved in full precision throughout the SFMA operation before the final rounding.

Both positive and negative Inf values are supported in the output, i.e., with a1=a2=±maxfp32a_{1}=a_{2}=\pm\mathrm{max}_{\mathrm{fp32}}, which is (2−2−23)​2127(2-2^{-23})2^{127}, and b1=b2=1b_{1}=b_{2}=1, the output is ±\pmInf. However, when positive and negative infinities are added, the resulting NaN carries a negative sign i.e., a1=−a2=a_{1}=-a_{2}= Inf, b1=b2=1b_{1}=b_{2}=1, the output is −-NaN. Subnormal values are supported in both input and output. For example, when a1=±2−127a_{1}=\pm 2^{-127} and b1=1b_{1}=1, the output is d=±2−127d=\pm 2^{-127}. Similarly, when a1=±b1=2−65a_{1}=\pm b_{1}=2^{-65}, the output is d=±2−130d=\pm 2^{-130}, which is a subnormal value rather than zero.

4.1.2. CDNA1 fp16 input

In fp16 input format, the supported matrix sizes are 32x32x4, 16x16x4, 4x4x4, 32x32x8 and 16x16x16. We first determined that subnormals are supported both in input, because we obtained d=±2−48d=\pm 2^{-48} for a1=2−24a_{1}=2^{-24}, b1=±2−24b_{1}=\pm 2^{-24}, and also in the output by producing d=cd=c for c=±2ec=\pm 2^{e} for e≤−126e\leq-126. Then we have determined the final rounding mode to be RNE using the same test vector as used in the case of fp32 input format. The SFMA test produced d=1+2−22+2−23d=1+2^{-22}+2^{-23} for all permutations which reflects the absence of an SFMA structure. Moreover, we permuted 11, 2−23+2−242^{-23}+2^{-24} between cc and p1p_{1}, while pj=2−23+2−24p_{j}=2^{-23}+2^{-24} for j>1j>1 while pℓ=0​∀ℓ≠j&ℓ≠1p_{\ell}=0~\forall\ell\neq j\And\ell\neq 1. When jj was varied from 22 onward, we obtained d=1+2−22+2−23d=1+2^{-22}+2^{-23} for j≤4j\leq 4 and d=1+2−21d=1+2^{-21} for j>4j>4, an SFMA-like behaviour, which revealed that the NFMA=4{N_{\mathrm{FMA}}}=4 for all supported matrix sizes. This test differs from the NFMA{N_{\mathrm{FMA}}} detection algorithm in (A. Khattak and Mikaitis, 2025), as the final rounding mode in CDNA 1 is detected to be RNE instead of RTZ. The permutation aspect shows that cc is added along with the products, via simultaneous alignment of significands, revealing an NVIDIA tensor-cores like behaviour (Khattak and Mikaitis, 2026).

In multi-term floating point addition (Mikaitis, 2024) where significands are aligned w.r.t the largest exponent, we expect to see finite number of alignment bits beyond the output precision fractional bits. To determine this feature, we performed the following tests:

  1. (1)

    we permuted 11, 2−242^{-24}, 2−25−neab2^{-25-{n_{\mathrm{eab}}}} between cc, p1p_{1}, p2p_{2}, a similar approach as that of (A. Khattak and Mikaitis, 2025; Hickmann and Bradford, 2019), where neab{n_{\mathrm{eab}}} is varied from 0,…,240,\dots,24 because the minimum value of product exponent is −48-48 in fp16 input format, we found d=1+2−23d=1+2^{-23} for every value of neab{n_{\mathrm{eab}}}. This showed that either there are multiple sticky bits (Tenca, 2009), or a correctly rounded implementation that adds all products in full precision (a Kulisch accumulator (Kulisch, 2008)), is present.

  2. (2)

    by permuting 11, 2−23+2−242^{-23}+2^{-24}, −2−25−neab-2^{-25-{n_{\mathrm{eab}}}}, dd remains 1+2−231+2^{-23} for all possible values of neab≥0{n_{\mathrm{eab}}}\geq 0. The output remains unchanged even when c=−2−25−neabc=-2^{-25-{n_{\mathrm{eab}}}}, which is in fp32 format, and therefore, neab{n_{\mathrm{eab}}} is varied from 0,…,(149−25)0,\dots,(149-25) due to the subnormal support.

  3. (3)

    we permuted ±1\pm 1, ±2−23\pm 2^{-23}, ±∑ℓ=−25L2−ℓ\pm\sum_{\ell=-25}^{L}2^{-\ell}, ±2−L\pm 2^{-L} where LL is varied from −26-26 to −48-48, for product computed with two fp16 format elements. We obtained d=±(1+2−22)d=\pm(1+2^{-22}) for all cases. Note all elements are either positive or negative.

The results of these tests reveal full precision addition or correctly rounded accumulation, and therefore, theoretically we can say that neab=∞{n_{\mathrm{eab}}}=\infty. These results are contrary to that reported by (Li et al., 2024) that MI100 matrix cores have 33 extra alignment bits. Unlike the NVIDIA tensor cores, the denormalised-product feature is not relevant here because the accumulation is performed accurately. Consequently, this feature may not be directly testable, but it is also unnecessary, since neab=∞{n_{\mathrm{eab}}}=\infty. Both positive and negative infinities are detected in the input i.e. a1=±a_{1}=\pmInf, b1=1⇒d=±b_{1}=1\Rightarrow d=\pmInf, and in case of both positive and negative infinities, the output is NaN with a negative sign i.e. a1=−a2=a_{1}=-a_{2}= Inf, b1=b2=1⇒d=−b_{1}=b_{2}=1\Rightarrow d=-NAN. Within the fp16 input format, the product exponent can reach a minimum value of −28-28, and with subnormal inputs, it can be as low as −48-48. Therefore, this case cannot be verified if the product is rounded to fp32. However, for the bf16 input format, this verification has been performed, as discussed further below.

4.1.3. CDNA1 bf16 inputs

The supported matrix sizes are 32x32x2, 16x16x2, 4x4x2, 32x32x4, and 16x16x8. In all supported sizes, we were able to determine that NFMA=2{N_{\mathrm{FMA}}}=2 performing identical test as carried out for fp16 case. Similarly, subnormals both in input and output are also supported with final rounding as RNE. The product of two bf16 numbers has a significantly wider dynamic range than a single bf16 number, as their exponents add during multiplication, resulting in an approximate range from 2−2662^{-266} to 22542^{254}, therefore, we determine if products are kept in full precision instead of rounding to fp32 before addition. To determine this, we set p1=±(2−149+2−150)p_{1}=\pm(2^{-149}+2^{-150}), p2=±2−149p_{2}=\pm 2^{-149}, and then p1=±(2−148+2−150)p_{1}=\pm(2^{-148}+2^{-150}), p2=±2−149p_{2}=\pm 2^{-149}, with both products residing within one block FMA, we obtained d=±2−148d=\pm 2^{-148}, and d=±2−147d=\pm 2^{-147}, respectively, for both cases. This shows that products are kept in full precision as if it were not the case, and the products were rounded to fp32 format via RNE before addition, we would have obtained d=±(2−148+2−149)d=\pm(2^{-148}+2^{-149}) for both cases. For other types of rounding of the product to fp32 format, the output would still be different. In addition, we set c=±1c=\pm 1, p1=±(2−23+2−24)p_{1}=\pm(2^{-23}+2^{-24}), and p2=∓2−25−neabp_{2}=\mp 2^{-25-{n_{\mathrm{eab}}}}, where cc and p1p_{1} have opposite signs to p2p_{2}, and neab=0,…,(266−25){n_{\mathrm{eab}}}=0,\dots,(266-25), and we obtained d=±(1+2−23)d=\pm(1+2^{-23}) which further confirms that products are computed and stored as the sum of exponents multiplied by the product of significands in full precision.

Infinities are detected at the bf16 input stage, even when the corresponding product would not overflow in fp32 precision. For example, a1=±2128a_{1}=\pm 2^{128}, which represent infinity in bf16, and b1=2−20b_{1}=2^{-20} yields p1=±2108p_{1}=\pm 2^{108}, which is representable in fp32. Nevertheless, the computed result is d=±∞d=\pm\infty. However, if the inputs are not infinities, the intermediate products are allowed to grow beyond 21282^{128} without being mapped to infinities. For instance, we obtained d=0d=0 for p1=2129p_{1}=2^{129}, p2=−2129p_{2}=-2^{129}, generated via a1=265a_{1}=2^{65}, b1=264b_{1}=2^{64}, and a2=265a_{2}=2^{65}, b2=−264b_{2}=-2^{64}. This suggests that infinity detection occurs at the input level, whereas overflow during intermediate product accumulation is handled at the very end when sum is converted to fp32. On the opposite end of the dynamic range, where the product falls below the smallest normal fp32-representable magnitude, we observe from the output that d=±2−148d=\pm 2^{-148} for c=±2−149c=\pm 2^{-149}, p1=±∑ℓ=1511562−ℓp_{1}=\pm\sum_{\ell=151}^{156}2^{-\ell}, p2=±2−156p_{2}=\pm 2^{-156}. This indicates that products smaller than 2−1492^{-149} still participate in the computation and are not flushed to zero. Because CDNA 1 matrix cores manifest full-precision accumulator within the dot products, many of the features later tested on CDNA 2 and CDNA3, as well as those tested on NVIDIA tensor cores (Khattak and Mikaitis, 2026), do not apply. With both ++Inf and −-Inf in inputs, the output is −-NaN, consistent with fp32 and fp16 input formats.

4.1.4. CDNA1 summary

The entire matrix core block diagram for CDNA 1 arch is illustrated in Fig. 1, where multiplication blocks represent full precision products.

Refer to caption
Figure 1. CDNA 1 matrix core model diagram. (a) fp16 input format with an fma size of 44, (b) for bf16 input format with an fma size of 22, and (c) fp32 input format which is an SFMA. All products are in full precision, the sum is of full accuracy, and final result is rounded to fp32 via RNE.

4.2. CDNA 2 Architecture

CDNA 2 matrix cores have been investigated previously by (Xie et al., 2025), but without test vectors provided. Here we are providing exact test vectors and confirm the details of the matrix cores. We have observed behaviour in following input formats: The MFMA intrinsic instructions support the same matrix sizes as those in the CDNA1 architecture for fp32 inputs. All other features are identical to the CDNA1 matrix cores with fp32 input format.

4.2.1. CDNA2 fp16 input

The supported matrix sizes for fp16 are the same as those in the CDNA1 architecture. Subnormals are not supported neither in input nor in the output. This means that if products are zero, and cc is less than 2−1262^{-126}, output is zero. Similarly, for c=0,a1=2−24c=0,~a_{1}=2^{-24} and b1=1b_{1}=1, we obtained d=0d=0. The final rounding mode is detected to be RNE from d=±{1d=\pm\{1, 1+2−22}1+2^{-22}\} for c=±{2−24c=\pm\{2^{-24}, 2−23+2−24}2^{-23}+2^{-24}\}, p1=±1p_{1}=\pm 1. The addition of multi-term floating point mechanism is detected via the following tests:

  1. (1)

    permuting the values 1,2−23+2−241,~2^{-23}+2^{-24}, 2−23+2−242^{-23}+2^{-24} between cc, p1,p_{1}, and p2p_{2} produced different values of dd indicating absence of global alignment between these three mentioned terms.

  2. (2)

    with c=1c=1, p1=pj=2−23+2−24,pℓ=0p_{1}=p_{j}=2^{-23}+2^{-24},~p_{\ell}=0, where ℓ≠j&ℓ≠1\ell\neq j\And\ell\neq 1, for j<5j<5, we obtained d=1+2−22+2−23d=1+2^{-22}+2^{-23} while for j>4j>4, it produced d=1+2−21d=1+2^{-21} indicating fl⁡(c+fl⁡(p1+pj))\mathrm{fl}(c+\mathrm{fl}(p_{1}+p_{j})) for j<5j<5.

  3. (3)

    with p1=1p_{1}=1, p4=p3=2−23+2−24p_{4}=p_{3}=2^{-23}+2^{-24}, all remaining factors set to zero, we obtained d=1+2−22+2−23d=1+2^{-22}+2^{-23}. With p2=1,p4=p3=2−23+2−24p_{2}=1,p_{4}=p_{3}=2^{-23}+2^{-24} produced same results indicating fl⁡(pj+fl⁡(p3+p4)){{\mathrm{fl}}}(p_{j}+{{\mathrm{fl}}}(p_{3}+p_{4})) for j<3j<3.

  4. (4)

    lastly, p1=p3=1p_{1}=p_{3}=1, p2=p4=2−23+2−24p_{2}=p_{4}=2^{-23}+2^{-24} produced d=2+2−21d=2+2^{-21} confirming fl⁡(p1+p2)+fl⁡(p3+p4){{\mathrm{fl}}}(p_{1}+p_{2})+{{\mathrm{fl}}}(p_{3}+p_{4}) while additionally setting c=2−23+2−25c=2^{-23}+2^{-25} resulted in d=2+2−21+2−22d=2+2^{-21}+2^{-22} confirming a pair-wise sum in groups which follows computations: fl⁡(c+fl⁡(fl⁡(p1+p2)+fl⁡(p3+p4))){{\mathrm{fl}}}(c+{{\mathrm{fl}}}({{\mathrm{fl}}}(p_{1}+p_{2})+{{\mathrm{fl}}}(p_{3}+p_{4}))).

This shows that pair wise sum is computed in sizes of 44 which confirms the results reported by (Xie et al., 2025). Moreover, at each RNE operation, it is unclear whether a pairwise sum that becomes an fp32 subnormal value is preserved or flushed to zero. This behavior cannot be tested using fp16 inputs because such values cannot be represented, but it is confirmed later using bf16 inputs. The grouping size is 44 for all supported matrix sizes. Similarly, infinities are detected at the input level, and the output is NaN with a negative sign when both positive and negative infinities are present in the input, i.e. a1=a2=1,b1=−b2=a_{1}=a_{2}=1,b_{1}=-b_{2}= Inf, ⇒d=−\Rightarrow d=-NaN.

4.2.2. CDNA2 bf16 input

Unlike CDNA1, and also the fp16 input format in CDNA2, the bf16 format introduces an additional subscript _1k in the MFMA intrinsic instruction for certain input matrix sizes. The matrix sizes that include this additional subscript are 32x32x4, 16x16x4, 4x4x4, 32x32x8, and 16x16x16. The matrix sizes without this subscript are the same as those supported in the CDNA1 architecture. Similar to the fp16 case, the final rounding mode is RNE. The group-wise pair-sum test vectors listed for the fp16 case indicate a group size of 44 for the intrinsic instructions with the subscript _1k (see Figure 2(a)), whereas for the remaining instructions the group size is 22 as depicted in Figure 2(b).

Regarding the product precision, which was not possible in the fp16 input case, we evaluated the case p1=2128p_{1}=2^{128} and p2=−p1p_{2}=-p_{1}. The resulting output was d=−NaNd=-\mathrm{NaN}, indicating that the products are converted to fp32 before accumulation rather than being accumulated in full precision as in the CDNA 1 architecture. Subnormals are flushed after every fl​{⋅}{{\mathrm{fl}}}\{\cdot\} operation i.e. after each normalisation followed by RNE. For instance, a1=b2=2−64a_{1}=b_{2}=2^{-64}, a2=b1=2−63a_{2}=b_{1}=2^{-63}, which makes the product subnormal i.e., p1=p2=2−127p_{1}=p_{2}=2^{-127}, the output remains d=0d=0.

Similar to the fp16 case, both positive and negative Infs are detected in inputs aa, bb, and cc; for a1=1,b1=±a_{1}=1,~b_{1}=\pmInf, we obtain d=±d=\pmInf, whereas for a1=a2=1,b1=−b2=±a_{1}=a_{2}=1,~b_{1}=-b_{2}=\pmInf, we obtain d=−d=-NaN. Figure 2[b] illustrates the matrix core structure corresponding to the bf16 input format.

Refer to caption
Figure 2. CNDA 2 matrix core model diagram. (a) shows fp16 and bf16 (with subscript _1k) input format with a group size of 44, and (b) for bf16 input format, without _1k with a group size of 22. fp32 input format matrix cores are identical to the ones in the CDNA 1 architecture, and are not shown.

4.3. CDNA 3 Architecture

For the CDNA 3, we used MI300, MI300A and MI300X AMD GPUs for experiments. The features of matrix multipliers in the CDNA 3 are also reported by (Xie et al., 2025); however, no detailed explanation of how they were determined or test vectors are provided. Unlike CDNA 1 and 2, CDNA 3 provides support for fp8 inputs in the matrix cores.

In terms of fp32 input, the matrix sizes supported in CDNA 1 and CDNA 2 are also supported in the CDNA 3 behave as SFMAs.

4.3.1. CDNA3 fp16 input

The matrix sizes supported are the same as those in CDNA 1 and 2. Subnormal numbers are supported in both input and output, and the final rounding mode is RNE. On this architecture, we first tested for the existence of extra alignment bits by permuting ±1,±2−24\pm 1,~\pm 2^{-24}, and ±2−24\pm 2^{-24} between cc, p1,p_{1}, and p2p_{2}. The result d=±(1+2−23)d=\pm(1+2^{-23}) suggests presence of an extra alignment bit. This also at least indicates that cc is not added after normalization of the product sum. Otherwise, under the assumption of RNE rounding, the output would have been ±1\pm 1 for p1=±1p_{1}=\pm 1 and p2=c=±2−24p_{2}=c=\pm 2^{-24}. However, the stage at which cc is added to the product sum remains to be determined and is investigated through the additional tests presented further below.

Next, given neab≥1{n_{\mathrm{eab}}}\geq 1, we perform the FMA size detection test with c=1c=1, p1=2−24p_{1}=2^{-24}, pj=2−24p_{j}=2^{-24} and pi≠1&i≠j=0p_{i\neq 1\And i\neq j}=0. We obtain d=1+2−23d=1+2^{-23} for j≤8j\leq 8, indicating an NFMA=8{N_{\mathrm{FMA}}}=8, which confirms the results reported in (Xie et al., 2025). To rule out the possibility that odd- and even-indexed products are added separately i.e. inter-leaved pattern as detected in (Khattak and Mikaitis, 2026), we permuted the values ±1\pm 1, 00, and ±(2−24+2−25)\pm(2^{-24}+2^{-25}) across p1p_{1}, p2p_{2}, and p3p_{3}. The resulting outputs were invariant, consistently yielding d=±1d=\pm 1 which implies no interleaving. To confirm whether neab=1{n_{\mathrm{eab}}}=1 or neab>1{n_{\mathrm{eab}}}>1, we used the test vector p1=±1p_{1}=\pm 1, p2=±(2−24+2−24−j)p_{2}=\pm(2^{-24}+2^{-24-j}). For j>0j>0, we obtained d=±1d=\pm 1 indicating that bits beyond the 24th fractional bit are truncated during alignment. Since the final rounding mode is RNE, if bits beyond 2424th bit were not truncated, the output would not be ±1\pm 1. For instance, if bits beyond 2525th bit were truncated, dd would have been ±(1+2−23)\pm(1+2^{-23}). Moreover, if it was a round up or down, or RNE at the alignment, dd cannot be equal to one for both positive or negative signs. With a single extra alignment bit, we tested whether a feature similar to NVIDIA tensor cores exists, where products remain denormalized. Using p1=2.25p_{1}=2.25, p2=2−23p_{2}=2^{-23}, and p3=p4=2−24p_{3}=p_{4}=2^{-24}, we observed d=2.25+2−22d=2.25+2^{-22} for a1=b1=1.5a_{1}=b_{1}=1.5. In contrast, when a1=2.25a_{1}=2.25 and b1=1b_{1}=1, we obtained d=2.25d=2.25. This suggests that products are computed as the product of significands multiplied by 2(sum of exponents)2^{(\text{sum of exponents})} and are kept denormalised while being accumulated. To check whether cc is aligned and added together with the products, we set c=1c=1, p1=p2=2−25p_{1}=p_{2}=2^{-25}, and p3=2−24p_{3}=2^{-24}. We obtained d=1+2−23d=1+2^{-23}, indicating that cc is added after product accumulation; otherwise, we would have obtained d=cd=c because neab=1{n_{\mathrm{eab}}}=1 and final rounding mode is RNE. To further analyze the alignment of cc, we used p1=±1p_{1}=\pm 1, c=±(2−24+2−24−j)c=\pm(2^{-24}+2^{-24-{j}}) where j>0j>0. We observed:

  1. (1)

    for positive values, d=1d=1 for j>0j>0 indicates the truncation of bits beyond the 24th fractional bit when the significand of cc, denoted with scs_{c}, is shifted w.r.t the products’ significands sum, denoted as Spi,sumS_{p_{i},\mathrm{sum}}. Additionally, using p1=1p_{1}=1, p2=2−24p_{2}=2^{-24}, and c=2−24c=2^{-24}, we obtain d=1+2−23d=1+2^{-23}, confirming that the product sum is not rounded to output precision before cc is added.

  2. (2)

    on the negative axis, we obtained d=−(1+2−23)d=-(1+2^{-23}) for all j>0j>0, indicating that the shifted scs_{c} is rounded to 2424 fractional bits using RD.

These results are consistent with the observations reported in (Xie et al., 2025), although explicit test vectors are not provided in that work. The fractional bit width when products sum, which is not normalized before adding to cc, is shifted against cc, we performed following tests:

  1. (1)

    c=±1c=\pm 1, p1=±(2−24+2−24−j)p_{1}=\pm(2^{-24}+2^{-24-j}), with j>0j>0. We observed d=1+2−23d=1+2^{-23} for j<8j<8 and d=1d=1 for j≥8j\geq 8 in the positive case, and d=−1−2−23d=-1-2^{-23} for all j>0j>0 in the negative case. In isolation, these results appear to suggest that the shifted Spi,sum{S_{p_{i},\mathrm{sum}}} is rounded down to 3131 fractional bits.

  2. (2)

    Due to mismatches in randomized testing discussed in Section 5.1, we performed an addition test with c=1c=1, p1+⋯+pNFMA=−(2−24+∑ℓ=2630+j2−ℓ)p_{1}+\dots+p_{{N_{\mathrm{FMA}}}}=-(2^{-24}+\sum_{\ell=26}^{30+j}2^{-\ell}), where jj is varied from 00 to 1010. We obtained d=1−2−24d=1-2^{-24} for j≤2j\leq 2 and d=1−2−23d=1-2^{-23} for j>2j>2. This test reflects that the shifted Spi,sum{S_{p_{i},\mathrm{sum}}} is rounded-down to 3232 fractional bits instead to 3131 fractional bits.

  3. (3)

    To further validate this conclusion and resolve the apparent contradiction with the previous test, we set c=1c=1, and chose p1+⋯+pNFMA=−(2−24+∑ℓ=26312−ℓ)+2−33p_{1}+\dots+p_{{N_{\mathrm{FMA}}}}=-(2^{-24}+\sum_{\ell=26}^{31}2^{-\ell})+2^{-33} where the 3232nd bit is intentionally set to 00 to prevent the carry generated by RD from propagating to the 2525th fractional bit. We obtained d=1−2−24d=1-2^{-24}, confirming that alignment-stage RD is applied at the 3232nd fractional bit of the shifted Spi,sum{S_{p_{i},\mathrm{sum}}}. Assuming this is the case, the sum of the aligned scs_{c} and Spi,sum{S_{p_{i},\mathrm{sum}}}, denoted by SaccS_{\mathrm{acc}}, is Sacc=1−2−24−∑ℓ=26322−ℓS_{\mathrm{acc}}=1-2^{-24}-\sum_{\ell=26}^{32}2^{-\ell}, which results in d=1−2−24d=1-2^{-24} after RNE. In contrast, if RD were applied at the 3131st fractional bit, then Sacc=1−2−24−∑ℓ=26312−ℓ−2−31=1−2−24−2−25S_{\mathrm{acc}}=1-2^{-24}-\sum_{\ell=26}^{31}2^{-\ell}-2^{-31}=1-2^{-24}-2^{-25} which would produce d=1−2−23d=1-2^{-23} after RNE, contrary to the observed result.

The observation that the shifted Spi,sum{S_{p_{i},\mathrm{sum}}} is rounded down to 32 fractional bits contradicts the 31 fractional bits reported in (Xie et al., 2025), yet it surprisingly matches the behaviour implemented in the associated repository. The observation from earlier test that d=1+2−23d=1+2^{-23} for j<8j<8 and d=1d=1 for j≥8j\geq 8 cannot be attributed to a 3131-bit alignment precision. With alignment-stage RD confirmed to retain 3232 fractional bits, the apparent 3131-bit behaviour instead arises from a subsequent RD operation applied to the normalised accumulated significand SaccS_{\mathrm{acc}} before the final RNE step. This observation was made due to mismatch in the randomized testing.

To characterize this second RD operation, we investigate whether SaccS_{\mathrm{acc}} is truncated or rounded to a 3232-bit representation before the final RNE stage by executing the following tests:

  1. (1)

    c=2−2−22c=2-2^{-22}, p1=2−22p_{1}=2^{-22}, and p2=2−23+2−jp_{2}=2^{-23}+2^{-j}, where jj is varied from 2424 to 3232. This produces d=2+2−22d=2+2^{-22} for j≤30j\leq 30 and d=2d=2 for j>30j>30, seemingly suggesting that the RD at the alignment stage operates at the 3030th fractional bit. However, with the RD at the 3232nd bit confirmed, this behavior strongly indicates the presence of an additional rounding operation on SaccS_{\mathrm{acc}} before the final RNE. For j=31j=31 and 3232, applying RD at the 3232nd bit yields Spi,sum=2−2−23+2−j{S_{p_{i},\mathrm{sum}}}=2-2^{-23}+2^{-j}. If SaccS_{\mathrm{acc}} is truncated to a total of 3232 bits at this stage, the resulting values are consistent with the observed results.

  2. (2)

    Additionally, we set c=−2+2−21c=-2+2^{-21}, p1=−2−21p_{1}=-2^{-21}, p2=−2−22p_{2}=-2^{-22}, and p3=−∑ℓ=24j2−ℓp_{3}=-\sum_{\ell=24}^{j}2^{-\ell}, where jj is varied from 2828 to 3232. We obtain d=−2−2−22d=-2-2^{-22} for j=28,…,30j=28,\ldots,30 and d=−2−2−21d=-2-2^{-21} for j>30j>30. It further confirms that the alignment stage RD occurs at the 3232nd fractional bit after Spi,sum{S_{p_{i},\mathrm{sum}}} is shifted. In addition, it also suggests that SaccS_{\mathrm{acc}} is rounded down, rather than simply truncated to 32 bits. For example, consider the case j=31j=31. Assuming the shifted Spi,sum{S_{p_{i},\mathrm{sum}}} is rounded down to 3232 fractional bits, the SaccS_{\mathrm{acc}} would be −2−2−22−∑ℓ=24312−ℓ-2-2^{-22}-\sum_{\ell=24}^{31}2^{-\ell}, which has 2 integer and 32 fractional bits. Truncation of SaccS_{\mathrm{acc}} to 3232 bits would result −2−2−22−∑ℓ=24302−ℓ-2-2^{-22}-\sum_{\ell=24}^{30}2^{-\ell}, and RNE, d=−2−2−22d=-2-2^{-22}. However, the observed result is d=−2−2−21d=-2-2^{-21} which indicates that the final sum of significands is rounded down to 32 bits rather than simply truncated.

In both tests above, SaccS_{\mathrm{acc}} has 22 integer bits and 3232 fractional bits i.e.,

Sacc=2+2−23+2−31=10\displaystyle S_{\mathrm{acc}}=2+2^{-23}+2^{-31}=10 .00000000000000000000001000000010​Test#1 for j=31\displaystyle.00000000000000000000001000000010~\text{Test\#1 for $j=31$}
Sacc=−2−2−22−∑ℓ=24312−ℓ=−10\displaystyle S_{\mathrm{acc}}=-2-2^{-22}-\sum_{\ell=24}^{31}2^{-\ell}=-10 .00000000000000000000001011111110​Test# 2 for ​j=31\displaystyle.00000000000000000000001011111110~\text{Test\# 2 for }j=31

However, the post-RD operation on SaccS_{\mathrm{acc}} appears to retain only 3030 fractional bits. In the positive case, the last 2 bits are discarded, whereas in the negative case the result is rounded down to −2−2−22−2−23-2-2^{-22}-2^{-23}. Now consider an earlier test with c=1c=1, p1=2−24p_{1}=2^{-24}, and p2=2−jp_{2}=2^{-j}. This test produced d=1+2−23d=1+2^{-23} for j=31j=31 and d=1d=1 for j=32j=32. In this case,

Sacc\displaystyle S_{\mathrm{acc}} =1+2−24+2−31=1.00000000000000000000000100000010\displaystyle=1+2^{-24}+2^{-31}=1.00000000000000000000000100000010
Sacc\displaystyle S_{\mathrm{acc}} =1+2−24+2−32=1.00000000000000000000000100000001\displaystyle=1+2^{-24}+2^{-32}=1.00000000000000000000000100000001

which clearly requires 11 integer bit and 3131 and 3232 fractional bits, respectively. The RD operation appears to preserve 3131 fractional bits. This behaviour suggests that the RD operation applied to SaccS_{\mathrm{acc}} may be accompanied by a small normalization step. To confirm this, we set c=1c=1 and p1+⋯+pNFMA=−(2−24+∑ℓ=26322−ℓ)p_{1}+\dots+p_{N_{\mathrm{FMA}}}=-(2^{-24}+\sum_{\ell=26}^{32}2^{-\ell}) which yields in

Sacc=1−2−24−∑ℓ=26322−ℓ=0.111111111111111111111110⏟24 bits​10000001.\displaystyle S_{\mathrm{acc}}=1-2^{-24}-\sum_{\ell=26}^{32}2^{-\ell}=0.\underbrace{111111111111111111111110}_{\text{24 bits}}10000001.

If the RD operation on SaccS_{\mathrm{acc}} were performed blindly, i.e., without a normalisation, then the last 11 bit would have to be discarded, as in the earlier case where 3131 fractional bits remained. Consequently, we would obtain d=1−2−23d=1-2^{-23}. However, since the actual value of dd is 1−2−241-2^{-24} which shows the RNE operation has not rounded up the output, confirms that the RD operation is applied after the normalisation of SaccS_{\mathrm{acc}}, and therefore not discarding the last bit. This constitutes the main normalisation step. In addition, two smaller normalisation steps are performed: one before and one after the final RNE operation. We cannot test this intermediate normalisation before RD operation of SaccS_{\mathrm{acc}} for a subnormal cc in this format, as the minimum possible product exponent is only −28≫−126-28\gg-126.

Lastly, since the products are accumulated first and cc is accumulated afterward, we consider a cancellation test in which the Spi,sum{S_{p_{i},\mathrm{sum}}} becomes zero, but the maximum product exponent exceeds that of cc by more than 24, i.e., c=210c=2^{10} and p1=−p2=210+25p_{1}=-p_{2}=2^{10+25}. It produces d=0d=0, indicating that cc is aligned w.r.t the maximum product exponent, even if Spi,sum{S_{p_{i},\mathrm{sum}}}.

4.3.2. CDNA3 bf16 input

All features are the same as that in fp16 input format. The product overflow feature could not have been validated in fp16 due to narrow range. However, in bf16, even though products are computed as product of significands times sum of the exponents raised to the base 22, we obtained ±\pmInf for |p1|≥2128|p_{1}|\geq 2^{128}. Moreover, since the products are kept denormalised, for which we suspected to obtain d=0d=0 for a1=a2=1.5×264a_{1}=a_{2}=1.5\times 2^{64}, b1=−b2=1.5×263b_{1}=-b_{2}=1.5\times 2^{63}, but it produced d=−d=-NaN which shows that even though the denormalised product’s, 2.25×21272.25\times 2^{127}, exponent is below 128128, the overflow is detected based on the magnitude instead of only on the exponent. However, an intermediate sum is allowed to exceed 21282^{128} within one block FMA operation. We further investigated whether a feature analogous to the exponent limitation observed in NVIDIA GPUs (Khattak and Mikaitis, 2026) is present. To this end, we conducted a series of tests by setting c=0c=0, p1=±2−150p_{1}=\pm 2^{-150}, and p2=±2−152−ip_{2}=\pm 2^{-152-i}, where ii is varied from 00 to 2323, corresponding to the 2424 fractional bits. The following observations are made:

  1. (1)

    on the positive axis, it produced d=2−149d=2^{-149} for i<6i<6, otherwise d=0d=0. This suggests that the minimum exponent used for aligning the significands, as reported in NVIDIA tensor core (Khattak and Mikaitis, 2026) is −132-132.

  2. (2)

    on the negative axis, we obtained d=−2−149d=-2^{-149} for all values of i≥0i\geq 0. This indicates that the product sum p1+⋯+pNFMA=−2−150−2−152−ip_{1}+\dots+p_{{N_{\mathrm{FMA}}}}=-2^{-150}-2^{-152-i} is not directly rounded to the output precision as otherwise we would have obtained identical results to the above case but in negative; instead, it is first aligned w.r.t the maximum exponent of −126-126, even though c=0c=0. This alignment yields d=−2−149d=-2^{-149} for i<6i<6. For i>5i>5, terms such as −2−158-2^{-158} or smaller fall beyond the 3131st fractional bit after alignment, and the product sum gets rounded down, which consistently results in d=−2−149d=-2^{-149}.

  3. (3)

    The exponent limitation is not observed for zero-valued products. Specifically, when p=1×0=0p=1\times 0=0, the behaviour is only restricted to c=0c=0.

From these observations, we can confirm c=0c=0 is considered as subnormal number as its exponent is considered to be −126-126. Unlike NVIDIA GPU tensor cores, where c=0c=0 is considered as a special case (Khattak and Mikaitis, 2026). This feature of exponent of cc, when it is zero, being -126, is not discussed by (Xie et al., 2025). However, this behavior is correctly implemented in the associated GitHub repository.

The normalisation step prior to RD application to SaccS_{\mathrm{acc}} when cc is a subnormal value could not be tested in fp16 input case. Therefore, for this input format, we set c=2−127−2−140c=2^{-127}-2^{-140} while p1+⋯+pNFMA=2−140+2−150+2−154−jp_{1}+\dots+p_{{N_{\mathrm{FMA}}}}=2^{-140}+2^{-150}+2^{-154-j} where jj is varied from 00 to 1010. For j≤3j\leq 3, we obtain d=2−127+2−149d=2^{-127}+2^{-149}, while for j>3j>3, d=2−127d=2^{-127}. Since the maximum alignment exponent is −126-126, SaccS_{\mathrm{acc}} has a zero integer bit; nevertheless, 31 fractional bits are retained, as −126−(−154−j≤3)=31-126-(-154-j\leq 3)=31. Thus, the intermediate normalisation before RD operations is also subnormal aware.

4.3.3. XF32 (tf19)

In CDNA 3 additional MFMA instructions are available: f32_16x16x8_xf32 and f32_16x16x4_xf32. These instructions accept float inputs but internally reduce the mantissa to 1010 bits. This behaviour was verified experimentally by setting a1=1+2−10+2−11a_{1}=1+2^{-10}+2^{-11} and b1=±1b_{1}=\pm 1, which produced d=±(1+2−10)d=\pm(1+2^{-10}), reflecting truncation, instead of ±(1+2−9)\pm(1+2^{-9}) which would reflect an application of RNE. The NFMA=4{N_{\mathrm{FMA}}}=4 in this input format, which matches with that of Ampere and Ada family GPUs tensor cores while that for Hopper and Blackwell, it is 8. We also confirm that when c=0c=0 the exponent of cc is equal to −126-126 in this input format as well.

4.3.4. CDNA3 binary8 input

In the binary8 input format, input matrices can be in any combination of E5M2 (fp8) and E4M3 (bf8), with matrix sizes 32x32x16 and 16x16x32. Subnormals are supported at both input and output, and the final rounding mode is RNE. The detected FMA size is 1616.

Permuting ±1\pm 1, 00, and ±(2−24+2−25)\pm(2^{-24}+2^{-25}) among p1p_{1}, p2p_{2}, and p3p_{3} yields d=±1d=\pm 1 when p2=0p_{2}=0. For other permutations, d=1d=1 for positive values and d=−(1+2−23)d=-(1+2^{-23}) for negative values. This confirms the interleaving feature reported in (Xie et al., 2025). These results indicate that even- and odd-indexed products are accumulated separately, with truncation applied beyond 2424 fractional bits (i.e., neab=1{n_{\mathrm{eab}}}=1). The resulting partial sums are then accumulated without intermediate normalization, and bits beyond 2424 fractional bits are rounded down.

To analyse the addition of cc, we permute the values ±1\pm 1 and ±(2−24+2−24−i)\pm\big(2^{-24}+2^{-24-i}\big) between cc and p1p_{1}, for i=0,…,10i=0,\dots,10. Similar behaviour to the fp16 and bf16 cases is observed: The shifted significand is rounded down to 24 fractional bits when cc is shifted, and to 32 fractional bits when the accumulated odd-indexed product sum is shifted. Repeating this experiment with permutations between cc and p2p_{2} yields identical behaviour. To further investigate the treatment of cc when it is shifted entirely beyond the retained precision of the product sum, we fix p1=±1p_{1}=\pm 1 and p2=2−24p_{2}=2^{-24}, while varying c=±2−24−ic=\pm 2^{-24-i}. For i=1i=1, the output is d=1d=1 for positive inputs and d=−(1+2−23)d=-(1+2^{-23}) for negative inputs, consistent with the behaviour observed previously. However, for i>1i>1, the result becomes d=±1d=\pm 1 in both cases. This indicates that once the entire significand of cc is shifted beyond the 2525th fractional bit of the accumulated product sum, truncation is applied instead of RD. The contrasting behaviour is observed when p1=±1p_{1}=\pm 1, and c=±(2−24+2−25−i)c=\pm(2^{-24}+2^{-25-i}), with i≥0i\geq 0, for which the result is d=1d=1 for the positive and d=−1−2−23d=-1-2^{-23} for the negative case. The underlying reason is that part of the significand of cc remains within the first 2525-bit fractional bits of the product sum, allowing the round-down effect to occur.

Finally, to determine whether odd- and even-indexed product sums are combined prior to adding cc, we set c=±(2−24+2−25)c=\pm(2^{-24}+2^{-25}) and permute ±1\pm 1 and ±2−25\pm 2^{-25} between p1p_{1} and p2p_{2}, emulating odd- and even-indexed partial sums. This produces d=1d=1 for positive values and d=−(1+2−22)d=-(1+2^{-22}) for negative values. This confirms that odd- and even-indexed product sums are first accumulated, followed by the addition of cc. If any alternate ordering were present, at least one permutation would yield d=1+2−23d=1+2^{-23} for positive values. We obtained identical results for the cancellation test mentioned in Section 4.3.1. However, the odd- and even-indexed product terms separate accumulation requires a unique test in which one of the product sums is zero but its corresponding exponent is the maximum. For example, we set c=2−20c=2^{-20}, a1=b1=20a_{1}=b_{1}=2^{0}, and a2=−a4=25a_{2}=-a_{4}=2^{5}, b2=b4=1b_{2}=b_{4}=1, which results in d=1d=1 instead of d=1+2−20d=1+2^{-20}. Although the even-indexed product sum is zero, the maximum exponent against which cc is aligned is 44, rather than 00. Thus, the maximum exponent is used for alignment regardless of whether the corresponding significand sum is zero.

4.3.5. CDNA3 summary

The CDNA 3 architecture’s matrix core model is depicted in Fig. 3. The block diagram is supported by Algorithm 1 and Algorithm 2 for clarity. The binary32 input model corresponds to an SFMA and is therefore not shown; however, it is supported in the CDNA 3 architecture.

Algorithm 1 CDNA 3 fp16/bf16/tf32 Input Formats Accumulation
1: pip_{i} for i=1,…,NFMAi=1,\dots,{N_{\mathrm{FMA}}}, and cc
2: dd
3: emax=max⁡([ep1,…,ep1])e_{\mathrm{max}}=\mathrm{max}([e_{p_{1}},\dots,e_{p_{1}}])
4: align spi|1,…,NFMAs_{p_{i}|1,\dots,{N_{\mathrm{FMA}}}} to emaxe_{\max}, truncate to 2424 fractional bits, accumulate into Spi,sum{S_{p_{i},\mathrm{sum}}} 
5: ec=−126e_{c}=-126 if c=0c=0
6: if emax≥ece_{\mathrm{max}}\geq e_{c} then
7:   sc′←s_{c}^{\prime}\leftarrow shift scs_{c} to right by (emax−ec)(e_{\mathrm{max}}-e_{c}) and RD it to 2424 frac. bits
8:   Sacc=Spi,sum+sc′S_{\mathrm{acc}}={S_{p_{i},\mathrm{sum}}}+s_{c}^{\prime} 
9: else
10:   Spi,sum′←{S^{\prime}_{p_{i},\mathrm{sum}}}\leftarrow shift Spi,sumS_{p_{i},\mathrm{sum}} to right by (ec−emax)(e_{c}-e_{\mathrm{max}}) and RD it to 3232 frac. bits.
11:   emax=ece_{\mathrm{max}}=e_{c}; Sacc=Spi,sum′+scS_{\mathrm{acc}}={S^{\prime}_{p_{i},\mathrm{sum}}}+s_{c} 
12:   {Sacc,emax}←\{S_{\mathrm{acc}},e_{\mathrm{max}}\}\leftarrow normalise {Sacc,eexp}%\{S_{\mathrm{acc}},e_{\mathrm{exp}}\}~\%subnormal-aware normalisation 
13:   Sacc←S_{\mathrm{acc}}\leftarrowRD SaccS_{\mathrm{acc}} to 31 frac. bits
14: end if
15: d←d\leftarrow convert {Sacc,emax}\{S_{\mathrm{acc}},e_{\mathrm{max}}\} to fp32 with RNE
Algorithm 2 CDNA 3 Binary8 Input Format Accumulation
1: pip_{i} for i=1,…,NFMAi=1,\dots,{N_{\mathrm{FMA}}}, and cc
2: dd
3: emax−odd=max⁡([ep1,ep3,…,epNFMA−1]),emax−even=max⁡([ep2,ep4,…,epNFMA])e_{\mathrm{max-odd}}=\mathrm{max}([e_{p_{1}},e_{p_{3}},\dots,e_{p_{N_{\mathrm{FMA}}}-1}]),~e_{\mathrm{max-even}}=\mathrm{max}([e_{p_{2}},e_{p_{4}},\dots,e_{p_{N_{\mathrm{FMA}}}}]) 
4: align odd and even indexed spis_{p_{i}} w.r.t emax−odde_{\mathrm{max-odd}} and emax−evene_{\mathrm{max-even}}, respectively 
5: truncate to 2424 fractional bits and accumulate into spi,sum−odd,s_{p_{i},\mathrm{sum-odd}}, and Spi,sum−evenS_{p_{i},\mathrm{sum-even}} 
6: emax=max⁡{emax−odd,emax−even}e_{\max}=\max\{e_{\mathrm{max-odd}},e_{\mathrm{max-even}}\} 
7: {Spi,sum−odd′,Spi,sum−even′}←\{S^{\prime}_{p_{i},\mathrm{sum-odd}},S^{\prime}_{p_{i},\mathrm{sum-even}}\}\leftarrow right shift {Spi,sum−odd,Spi,sum−even}\{S_{p_{i},\mathrm{sum-odd}},S_{p_{i},\mathrm{sum-even}}\} by {emax−emax−odd,emax−emax−even}\{e_{\mathrm{max}}-e_{\mathrm{max-odd}},e_{\mathrm{max}}-e_{\mathrm{max-even}}\}, respectively, and RD both to 2424 frac. 
8: Spi,sum=Spi,sum−odd′+Spi,sum−even′{S_{p_{i},\mathrm{sum}}}=S^{\prime}_{p_{i},\mathrm{sum-odd}}+S^{\prime}_{p_{i},\mathrm{sum-even}} 
9: ec=−126e_{c}=-126 if c=0c=0
10: if emax≥ece_{\mathrm{max}}\geq e_{c} then
11:   if emax−ec≤25e_{\mathrm{max}}-e_{c}\leq 25 then
12:    sc′←s_{c}^{\prime}\leftarrow shift scs_{c} to right by (emax−ec)(e_{\mathrm{max}}-e_{c}) and RD it to 2424 frac. bits
13:   else
14:    sc′=0s^{\prime}_{c}=0
15:   end if
16:   Sacc=Spi,sum+sc′S_{\mathrm{acc}}={S_{p_{i},\mathrm{sum}}}+s^{\prime}_{c}
17: else
18:   Follow step 8-11 of Algorithm 1
19: end if
20: d←d\leftarrow convert {Sacc,emax}\{S_{\mathrm{acc}},e_{\mathrm{max}}\} to fp32 with RNE
Refer to caption
Figure 3. Matrix core model diagram in CDNA 3. (a) fp16 and bf16 input format, (b) tf19 input format, and (c) binary8 input format.

5. Randomized Tests Suite and Refinement of Models

As our focus is on a single element of the output matrix DD, the randomized testing procedure compares the result of (1) for independently generated vectors aa, bb, and scalar cc in each run. While it makes sampling a lot easier, it does not make the sampling space small enough to be exhaustively searched. Where there can be many ways to sample the input space, we propose following ones with appropriate justification. With NensN_{\mathrm{ens}} denoting the finite ensemble size, and kk the length of the input vectors aa and bb which denotes the inner dimension supported by an MFMA intrinsic instruction, we perform all testing and experiments using the maximum supported value of kk.

In the verification process, we first generate a set of random inputs to evaluate the GPU matrix cores, with the outputs written to a text file. The same random inputs are then generated in MATLAB to evaluate the corresponding matrix-core models, followed by a comparison between the GPU and model outputs. We perform this verification on the MI100, MI210 and MI250 for fp16 and bf16 input formats, and on the MI300A/X (CDNA 3) for Binary8 (E5M2 and E4M3), fp16, bf16, and tf19. In total, we perform nine verification tests in two sampling strategies.

5.1. Exponent-Mantissa Sampling

A straightforward approach would be to draw a,b∼𝒰a,b\sim\mathcal{U}, where 𝒰\mathcal{U} denotes the uniform distribution. However, such sampling does not yield a uniform distribution of floating-point exponents. Rather, the exponent distribution is biased towards larger values. To obtain a more uniform coverage of the floating-point exponent range, we generate the random vectors by sampling the significand uniformly in the range [−2+2−f+1,2−2−f+1][-2+2^{-f+1},2-2^{-f+1}], where ff denotes the precision of the considered input format, as defined in Table 1. The corresponding exponents are drawn from a uniform integer distribution over the prescribed lower and upper bounds of the exponent range. The pseudo-code is outlined in Algorithm 3. With this sampling strategy, all architectures considered are tested on the basis of the input format exponent ranges described below.

Algorithm 3 Ensemble Generation Psuedo-code for Randomized Testing of Tensor Core Models
Choose kk and ff based on the input format and GPU architecture.
mt19937 rng(0)  %\%~Mersenne Twister with seed 0
choose Nens,ea−,ea+,eb−,eb+,ec−,ec+N_{\mathrm{ens}},~e_{\mathrm{a}}^{-},~e_{\mathrm{a}}^{+},e_{\mathrm{b}}^{-},~e_{\mathrm{b}}^{+},e^{-}_{\mathrm{c}},~e_{\mathrm{c}}^{+}
for i=1:Nensi=1:N_{\mathrm{ens}} do
  aℓ=sℓ​2eℓ,eℓ​∼i.i.d.​𝒰​(ea−,ea+),sℓ​∼i.i.d.​𝒰​(−2+2−f+1, 2−2−f+1),ℓ=1,…,ka_{\ell}=s_{\ell}2^{e_{\ell}},\quad e_{\ell}\overset{\mathrm{i.i.d.}}{\sim}\mathcal{U}(e_{\mathrm{a}}^{-},e_{\mathrm{a}}^{+}),\quad s_{\ell}\overset{\mathrm{i.i.d.}}{\sim}\mathcal{U}(-2+2^{-f+1},\,2-2^{-f+1}),\quad\ell=1,\ldots,k
  bℓ=sℓ​2eℓ,eℓ​∼i.i.d.​𝒰​(eb−,eb+),sℓ​∼i.i.d.​𝒰​(−2+2−f+1, 2−2−f+1),ℓ=1,…,k.b_{\ell}=s_{\ell}2^{e_{\ell}},\quad e_{\ell}\overset{\mathrm{i.i.d.}}{\sim}\mathcal{U}(e_{\mathrm{b}}^{-},e_{\mathrm{b}}^{+}),\quad s_{\ell}\overset{\mathrm{i.i.d.}}{\sim}\mathcal{U}(-2+2^{-f+1},\,2-2^{-f+1}),\quad\ell=1,\ldots,k.
  c=s​2e,c​∼i.i.d.​𝒰​(ec−,ec+),s​∼i.i.d.​𝒰​(−2+2−23, 2−2−23)c=s2^{e},\quad c\overset{\mathrm{i.i.d.}}{\sim}\mathcal{U}(e_{\mathrm{c}}^{-},e_{\mathrm{c}}^{+}),\quad s\overset{\mathrm{i.i.d.}}{\sim}\mathcal{U}(-2+2^{-23},\,2-2^{-23}) 
end for

5.1.1. Half Precision (fp16)

In this input format, all architectures are tested over Nest=106N_{\mathrm{est}}=10^{6} random inputs each with k=16k=16 and following exponent limits:

  • •

    ea−=eb−=−14e_{\mathrm{a}}^{-}=e_{\mathrm{b}}^{-}=-14, ec−=−126e_{\mathrm{c}}^{-}=-126, ea+=eb+=15e_{\mathrm{a}}^{+}=e_{\mathrm{b}}^{+}=15, and ec+=127e_{\mathrm{c}}^{+}=127.

  • •

    ea−=eb−=−24e_{\mathrm{a}}^{-}=e_{\mathrm{b}}^{-}=-24, ec−=−149e_{\mathrm{c}}^{-}=-149, ea+=eb+=−14e_{\mathrm{a}}^{+}=e_{\mathrm{b}}^{+}=-14, ec+=−126e_{\mathrm{c}}^{+}=-126. This tests the models in subnormal range where the exponent limits are set between minimum normal and minimum subnormal exponents.

5.1.2. BrainFloat16 (bf16)

Analogously to the FP16 input-format case, we use Nens=106N_{\mathrm{ens}}=10^{6} and k=16k=16 for CDNA 2 and 3, while for CDNA 1, kk is limited to a maximum value of 88, with the exponent limits set as given below:

  • •

    ea−=eb−=ec−=−126e_{\mathrm{a}}^{-}=e_{\mathrm{b}}^{-}=e_{\mathrm{c}}^{-}=-126, ea+=eb+=63e_{\mathrm{a}}^{+}=e_{\mathrm{b}}^{+}=63, ec+=127e_{\mathrm{c}}^{+}=127. The maximum exponent limit for aa and bb is set to 6363 to prevent the product from becoming infinite, as CDNA 3 maps |p|≥2128|p|\geq 2^{128} to infinity. This setting is adopted to evaluate the model using normal inputs, while special cases such as infinities and NaNs are treated separately, as they do not provide meaningful practical validation.

  • •

    ea−=eb−=−133e_{\mathrm{a}}^{-}=e_{\mathrm{b}}^{-}=-133, ec−=−149e_{\mathrm{c}}^{-}=-149, ea+=eb+=ec+=−126e_{\mathrm{a}}^{+}=e_{\mathrm{b}}^{+}=e_{\mathrm{c}}^{+}=-126 which tests the models between minimum normal and subnormal exponent limits for bf16 input format.

5.1.3. TensorFloat32 (tf19)

This format is supported on CDNA 3 and subsequent CDNA architectures, and is denoted as XF32 in the MFMA intrinsic. With k=8k=8 and Nens=106N_{\mathrm{ens}}=10^{6}, the same exponent limits as for bf16 were used; however, the minimum subnormal exponent was set to −136-136. We found zero mismatches.

5.1.4. Binary8 (fp8-e5m2, fp8-e4m3)

CDNA 3 supports this format as an input format for matrix core operations. The AMD intrinsic instructions for MFMA in 8-bit input formats, fp8-e5m2 and fp8-e4m3, are referred to as fp8 and bf8, respectively33 3 __builtin_amdgcn_mfma_f32_32x32x16_fp8_fp8, __builtin_amdgcn_mfma_f32_32x32x16_bf8_bf8. Also support mixed, i.e.,_fp8_bf8. Unlike the OCP fp8 formats, these are implemented using the FNUZ encoding. With Nens=106{N_{\mathrm{ens}}}=10^{6} and k=16k=16, following exponent limits are used.

  • •

    fp8/bf8: ea−=eb−=−7/−15e_{\mathrm{a}}^{-}=e_{\mathrm{b}}^{-}=-7/-15, ec−=−126e_{\mathrm{c}}^{-}=-126, ea+=eb+=7/15e_{\mathrm{a}}^{+}=e_{\mathrm{b}}^{+}=7/15, ec+=127e_{\mathrm{c}}^{+}=127.

  • •

    fp8/bf8: ea−=eb−=−10/−17e_{\mathrm{a}}^{-}=e_{\mathrm{b}}^{-}=-10/-17, ec−=−149e_{\mathrm{c}}^{-}=-149, ea+=eb+=−7/−15e_{\mathrm{a}}^{+}=e_{\mathrm{b}}^{+}=-7/-15, ec+=−126e_{\mathrm{c}}^{+}=-126.

For normal values, we scan the exponent range from the minimum to the maximum normal exponent. For subnormal values, we scan the range from the minimum normal exponent down to the minimum subnormal exponent. Minimum normal exponent in FNUZ is different than OCP encoding as outlined in Table 1. The CDNA 3 architecture is further evaluated at larger values of (NensN_{\mathrm{ens}}), with the underlying rationale discussed in the following subsections.

5.2. Independent Bit Sampling

This sampling strategy samples each bit independently from a Bernoulli distribution with equal probability. To avoid infinities in the input, which occur frequently for formats with small exponent ranges under this sampling strategy, we redraw the exponent whenever it results in an infinity.

5.3. Iterative Refinement Via Mismatch Based Hidden Feature Detection

Across all configurations considered above, no mismatches were observed between the matrix-core model and the GPU outputs for CDNA 1 or CDNA 2, indicating that the feature space was sufficient for both sampling strategies. For CDNA 3, the independent bit-sampling strategy likewise produced no mismatches. However, the first sampling strategy revealed a small number of mismatches during model refinement, as touched upon in Section 4.3.1. These mismatches occurred at different stages of the refinement process as successive updates were introduced. Due to the two-stage RD operation and the associated intermediate normalisation, the model required four refinement iterations in total, as summarised in Table 2. The mismatches were progressively eliminated through these successive refinements. In the fourth refinement iteration, the validation was extended to Nens=107{N_{\mathrm{ens}}}=10^{7} across the full exponent range using the first sampling strategy for fp16, bf16, and tf32, and to 2×1072\times 10^{7} for binary8. No mismatches were observed for any input format. For Nens=106{N_{\mathrm{ens}}}=10^{6}, each test required, on average, 13±1.513\pm 1.5 minutes on the MATLAB side and 11–22 minutes on the GPU.

5.4. Probabilistic Analysis

Suppose we are looking for triplets t=(a,b,c)t=(a,b,c), where aa,b∈ℝkb\in\mathbb{R}^{k}, in which the result from the matrix core operation in hardware differs from the one predicted by the software matrix core model. Suppose that the set of such triplets causing such mismatches is EE. Then, we would like to estimate the probability P⁡(t∈E|S)P(t\in E|S), that is, the probability, given a single triplet tt sampled according to SS, that tt produces a model error. We will refer to this probability for short as pSp_{S}.

Let’s suppose that we draw Nens{N_{\mathrm{ens}}} samples, and observe mm errors. The likelihood of this data is given by the binomial distribution:

L⁡(pS)=PB​i​n​(m|Nens,pS)=Nens!m!​(Nens−m)!​pSm​(1−pS)Nens−m.L(p_{S})=P_{Bin}(m|{N_{\mathrm{ens}}},p_{S})=\frac{{N_{\mathrm{ens}}}!}{m!({N_{\mathrm{ens}}}-m)!}p^{m}_{S}(1-p_{S})^{{N_{\mathrm{ens}}}-m}.

Assuming a uniform prior distribution density on pSp_{S}, i.e. π⁡(pS)=1\pi(p_{S})=1, 0≤pS≤10\leq p_{S}\leq 1, the posterior density for pSp_{S} is

π⁡(pS|Nens,m)=κ×π⁡(pS)×L⁡(pS)=(Nens+1)!m!​(Nens−m)!​pSm​(1−pS)Nens−m,\displaystyle\pi(p_{S}|{N_{\mathrm{ens}}},m)=\kappa\times\pi(p_{S})\times L(p_{S})=\frac{({N_{\mathrm{ens}}}+1)!}{m!({N_{\mathrm{ens}}}-m)!}p^{m}_{S}(1-p_{S})^{{N_{\mathrm{ens}}}-m},

which is a Beta distribution density on the 1-simplex. The ratio of factorials is a normalization constant guaranteeing that ∫01d​pS​π​(pS|Nens,m)=1\int^{1}_{0}dp_{S}\,\pi(p_{S}|{N_{\mathrm{ens}}},m)=1.

For the case m=0m=0, which is the usual case in which sampling has not turned up any errors, we have

π⁡(pS|Nens,m=0)=(Nens+1)×(1−pS)Nens.\pi(p_{S}|{N_{\mathrm{ens}}},m=0)=({N_{\mathrm{ens}}}+1)\times(1-p_{S})^{{N_{\mathrm{ens}}}}.

This density peaks at pS=0p_{S}=0, falling off towards 0 at pS=1p_{S}=1. We can construct an ϵ\epsilon-upper bound on pSp_{S} denoted by pϵp_{\epsilon}, which is defined by the equation

∫0pϵd​pS​π​(pS|Nens,m=0)=1−ϵ.\int^{p_{\epsilon}}_{0}dp_{S}\,\pi(p_{S}|{N_{\mathrm{ens}}},m=0)=1-\epsilon.

The quantity 1−ϵ1-\epsilon represents the level of certainty that pS<pϵp_{S}<p_{\epsilon}. Thus, we may choose ϵ\epsilon as small as we like (10−610^{-6}, say), and solve for the pϵp_{\epsilon} that produces this ϵ\epsilon.

The integral is

∫0pϵd​pS​π​(pS|Nens,m=0)\displaystyle\int^{p_{\epsilon}}_{0}dp_{S}\,\pi(p_{S}|{N_{\mathrm{ens}}},m=0) =\displaystyle= ∫0pϵd​pS​(Nens+1)×(1−pS)Nens\displaystyle\int^{p_{\epsilon}}_{0}dp_{S}\,({N_{\mathrm{ens}}}+1)\times(1-p_{S})^{{N_{\mathrm{ens}}}}
=\displaystyle= ∫0pϵd​pS​dd​ps​(1−pS)Nens+1=1−(1−pϵ)Nens+1.\displaystyle\int^{p_{\epsilon}}_{0}dp_{S}\,\frac{d}{dp_{s}}(1-p_{S})^{{N_{\mathrm{ens}}}+1}=1-(1-p_{\epsilon})^{{N_{\mathrm{ens}}}+1}.

Setting this equal to 1−ϵ1-\epsilon we obtain

pϵ\displaystyle p_{\epsilon} =\displaystyle= 1−ϵ1Nens+1=1−exp⁡(log⁡ϵNens+1).\displaystyle 1-\epsilon^{\frac{1}{{N_{\mathrm{ens}}}+1}}=1-\exp\left(\frac{\log\epsilon}{{N_{\mathrm{ens}}}+1}\right).

In the limit of many samples, such that Nens+1≫|log⁡ϵ|{N_{\mathrm{ens}}}+1\gg|\log\epsilon|, this becomes pϵ≈log⁡ϵ−1Nens+1p_{\epsilon}\approx\frac{\log\epsilon^{-1}}{{N_{\mathrm{ens}}}+1}.

We apply the probabilistic framework described above to quantify the confidence bounds of our experimental validation. The number of tested samples, observed mismatches, and the corresponding Bayesian upper bounds are summarized in Table 2 in each iteration of the refinement of the model. Thus, for fixed ϵ\epsilon, a tenfold increase in error-free trials reduces the upper bound by approximately tenfold. For ϵ=10−6\epsilon=10^{-6}, 10510^{5}, 10610^{6}, and 10710^{7} trials yield bounds of approximately 1.38×10−41.38\times 10^{-4}, 1.38×10−51.38\times 10^{-5}, and 1.38×10−61.38\times 10^{-6}, respectively. Hence, our choice of 10610^{6}–10710^{7} trials provides 10−510^{-5}–10−610^{-6}-level bounds while avoiding the proportionally greater cost of validation on the Matlab side.

Table 2. Experimental validation results across four model-refinement iterations and Bayesian upper bounds on the mismatch probability of the CDNA 3 matrix core model using exponent-mantissa sampling.
Iteration Input Format Nens{N_{\mathrm{ens}}} Mismatches(m)(m) Upper Bound on pϵ=10−6p_{\epsilon=10^{-6}}
0 binary8 (fp8, bf8) 2×1062\times 10^{6} 0 6.91×10−66.91\times 10^{-6}
fp16 10610^{6} 1 1.69×10−51.69\times 10^{-5}
bf16 10610^{6} 0 1.38×10−51.38\times 10^{-5}
tf32 10610^{6} 1 1.69×10−51.69\times 10^{-5}
1 binary8 (fp8, bf8) 2×1062\times 10^{6} 0 6.91×10−66.91\times 10^{-6}
fp16 10610^{6} 0 1.38×10−51.38\times 10^{-5}
bf16 10610^{6} 0 1.38×10−51.38\times 10^{-5}
tf32 10610^{6} 1 1.69×10−51.69\times 10^{-5}
2 binary8 (fp8, bf8) 2×1062\times 10^{6} 0 6.91×10−66.91\times 10^{-6}
fp16 10610^{6} 0 1.38×10−51.38\times 10^{-5}
bf16 10610^{6} 0 1.38×10−51.38\times 10^{-5}
tf32 10610^{6} 0 1.38×10−51.38\times 10^{-5}
3 fp16 1.25×1061.25\times 10^{6} 1 1.35×10−51.35\times 10^{-5}
4 binary8 (fp8, bf8) 2×1072\times 10^{7} 0 6.91×10−76.91\times 10^{-7}
fp16 10710^{7} 0 1.38×10−61.38\times 10^{-6}
bf16 10710^{7} 0 1.38×10−61.38\times 10^{-6}
tf32 10710^{7} 0 1.38×10−61.38\times 10^{-6}

6. Comparison of GPU Numerical Features

While NVIDIA GPU numerical features have been reported in (Khattak and Mikaitis, 2026; Xie et al., 2025), we briefly compare them with those of AMD GPUs in Table 3. The table summarizes key numerical features across several generations of NVIDIA tensor cores and AMD matrix cores, including product alignment, the FMA size, handling of the addend cc, alignment and rounding of the accumulated product sum, and the minimum supported exponent. These features highlight the differences in accumulation, rounding, and precision handling across GPU MMA architectures. The table speaks for itself—the approaches of evaluating dot products and matrix multiplies of floating-point numbers, in hardware, differ significantly between the two major companies. This translates to numerical error accumulation differences, as shown in the demonstrative applications of Section 8.

Table 3. Comparison of numerical features of several generations of NVIDIA tensor cores (mixed-precision matrix multipliers) (Khattak and Mikaitis, 2026) and AMD matrix cores. Newly derived models are in bold.
B200 H100 MI300A/X H100, B200 A100 MI100 MI300A/X H100, B200 A100 MI300A/X
In fp8 fp8 fp8 fp16/bf16 fp16/bf16 fp16/bf16 fp16/bf16 tf19 tf19 tf19
Prd. Align. (2,25) (2,13) (2,24) (2,25) (2,24) (1,∞\infty) (2,24) (2,25) (2,24) (2,24)
NFMAN_{\mathrm{FMA}} 32 32 8odd−even8_{\mathrm{odd-even}} 16/16 8/8 4/2 8/8 8 4 4
ec=0e_{c=0} −∞-\infty −∞-\infty −126-126 −∞-\infty −∞-\infty −∞-\infty −126-126 −∞-\infty −∞-\infty −126-126
cc late/early early early late early early early late early early late
align⁡(c,Spi−sum)\mathrm{align}(c,S_{p_{i}-\mathrm{sum}}). - - (24,32,RD) - - - (24,32,RD) - - (24,32,RD)
out. round. RTZ RTZ RNE RTZ RTZ RNE RNE RTZ RTZ RNE
eminlime^{\mathrm{lim}}_{\min} - - - -133 -132 - -126 -133 -132 -126
|pi||p_{i}|-overflow No No 21282^{128} No No No 21282^{128} No No 21282^{128}

Note: (24,32,RD)(24,32,\mathrm{RD}) indicates that shifted scs_{c} and Spi−sumS_{p_{i}-\mathrm{sum}} are rounded to 24 and 32 frac. bits during alignment, respectively; odd​-​even\mathrm{odd\text{-}even} denotes the interleaved product-addition pattern, where odd- and even-indexed products are accumulated separately. The feature ec=0e_{c=0} denotes the exponent of cc when c=0c=0, with ec=0=−∞e_{c=0}=-\infty indicating that cc is discarded during accumulation. MI210 and MI250 GPUs support fp16/bf16 and implement SFMA with the grouping shown in Fig. 2. The output is assumed to be in fp32 format. NVIDIA GPUs support fp16-format output for both fp8- and fp16-format inputs and implement eminlim={−20,−21}e_{\mathrm{min}}^{\mathrm{lim}}=\{-20,-21\} for A100 and H100/B200, respectively, across all GPUs, rather than −14-14, which is the minimum normal exponent of the fp16 format.

7. MFMA Software Models

This section introduces MATLAB Matrix Core v0.6, comprising the extensions to the software developed as part of this research. The toolbox was developed in MATLAB R2026a and uses CPFloat (Fasi and Mikaitis, 2023) for conversion from double- or single-precision formats to all supported input formats. The toolbox is structured around the generalised matrix multiplier model, which supports four main architectural configurations: correct_rounding, modelling a Kulisch-style correctly rounded accumulator such as CDNA 1; pair_wise_sum, modelling the pairwise accumulation used in CDNA 2; global_alignment, modelling the CDNA 3-like architecture for fp16, bf16, and tf32 inputs; and odd_even_grouping, modelling the CDNA 3-like architecture for binary8 inputs. Each configuration provides parameters to control the relevant architectural features, including NFMA{N_{\mathrm{FMA}}}, alignment and rounding behaviour, subnormal support, and exponent handling, thereby allowing both the AMD and alternative configurations to be simulated. With the Parallel Computing Toolbox available, matrix multiplications using the proposed models are parallelized across available CPU cores. The toolbox is also compatible with GNU Octave and can be accessed from Python via Oct2Py or the MATLAB Engine API. In future work, we will port the model back-end to C to improve performance. An example call of MI300A matrix core for a 4x4 matrix multiply accumulate is shown below.

⬇ inopts.format = ’fp8-e4m3’, outopts.format = ’binary32’; A = cpfloat(randn(4), inopts), B = cpfloat(randn(4), inopts); C = cpfloat(randn(4), outopts), alpha = 1, beta = 1; MI300AMC(alpha, A, B, beta, C, inopts.format, outopts.format) >> ans = 0.1484 -0.6631 -0.1836 -1.3271 0.8232 1.6418 0.4805 3.0227 3.6592 -0.1250 1.4902 2.2637 3.7432 -3.4275 0.2031 0.1663

8. Application Scenarios

In this section we apply the developed AMD matrix core models to several demonstration applications, acting as representative examples of how the models can be used. The source code of these experiments is available.44 4 https://github.com/north-numerical-computing/MATLAB-tensor-core/tree/main/experiments

8.1. Multi-word Arithmetic

We have implemented an algorithm for emulating high-precision matrix multiplication via a multi-word representation of matrices in one of the low-precision formats supported by the AMD matrix cores. The details are provided by Mary and Mikaitis (Mary and Mikaitis, 2025). We have plotted three generations of the AMD CDNA GPUs and one of NVIDIA, for comparison. In each diagram, the number of words determines the accuracy of the conversion of input matrices between binary64 representation and a NN-word low-precision representation: larger NN means lower error in converting the binary64 inputs.

Figure 4 shows the experiments with fp8-e5m2. The plot shows the relative norm-wise error as the inner dimension nn of two matrices being multiplied is increased. The main differences between the MI300X and the B200 appear when either the dimension nn is relatively large or when six words are used. This can be explained by the differences between these two devices as shown in Table 3. Figures 5 and 6 show similar experiments for fp16 and bf16. With these formats, CDNA 1 and 2 can also be utilised. Overall, all CDNA matrix multipliers yield similar error in this matrix multiplication algorithm, whilst B200 consistently has a larger error which grows with nn. This behavior can be attributed to the fact that NVIDIA tensor cores use RTZ, whereas AMD matrix cores use RNE for the final rounding. In addition, AMD applies RD when aligning the accumulation term with the product sum. However, the effect of this asymmetric rounding is not pronounced because the accumulation term often exceeds the inner-product term. Consequently, RD becomes effective only beyond the 31st fractional bit, which provides sufficient precision to reduce the impact of its asymmetric behavior on the final result.

10210^{2}10410^{4}10610^{6}10−910^{-9}10−510^{-5}10−110^{-1}nn‖C^−C‖∞‖A‖∞​‖B‖∞\frac{\|\widehat{C}-C\|_{\infty}}{\|A\|_{\infty}\|B\|_{\infty}}Single word (fp8-e5m2)10210^{2}10410^{4}10610^{6}10−910^{-9}10−510^{-5}10−110^{-1}nnQuad-word (fp8-e5m2)10210^{2}10410^{4}10610^{6}10−910^{-9}10−510^{-5}10−110^{-1}nn6-word (fp8-e5m2)
MI300XB200
Figure 4. Multi-word arithmetic experiment presented by Mary and Mikaitis (Mary and Mikaitis, 2025, Sec. 5) on the simulation of various tensor cores. We have reproduced the experiment on MI300X and NVIDIA B200 matrix multipliers modelled in MATLAB. Relative norm-wise errors of matrix multiplication, compared with a default MATLAB binary64 multiplication, are shown. The input matrices to the GEMM are A∈R10×nA\in R^{10\times n} and B∈Rn×10B\in R^{n\times 10}. These matrices are multiplied with a multi-word algorithm (Mary and Mikaitis, 2025, Sec. 4) by splitting them into several fp8-e5m2 words.
10210^{2}10410^{4}10610^{6}10−910^{-9}10−510^{-5}10−110^{-1}nn‖C^−C‖∞‖A‖∞​‖B‖∞\frac{\|\widehat{C}-C\|_{\infty}}{\|A\|_{\infty}\|B\|_{\infty}}Single word (binary16)10210^{2}10410^{4}10610^{6}10−910^{-9}10−510^{-5}10−110^{-1}nnDouble word (binary16)10210^{2}10410^{4}10610^{6}10−910^{-9}10−510^{-5}10−110^{-1}nnTriple word (binary16)
MI100MI250MI300XB200
Figure 5. Multi-word arithmetic experiment presented by Mary and Mikaitis (Mary and Mikaitis, 2025, Sec. 5) on the simulation of various tensor cores. Relative norm-wise errors of matrix multiplication, compared with a default MATLAB binary64 multiplication, are shown. The input matrices to the GEMM are A∈R10×nA\in R^{10\times n} and B∈Rn×10B\in R^{n\times 10}. These matrices are multiplied with a multi-word algorithm (Mary and Mikaitis, 2025, Sec. 4) by splitting them into several fp16 words.
10210^{2}10410^{4}10610^{6}10−910^{-9}10−510^{-5}10−110^{-1}nn‖C^−C‖∞‖A‖∞​‖B‖∞\frac{\|\widehat{C}-C\|_{\infty}}{\|A\|_{\infty}\|B\|_{\infty}}Single word (bfloat16)10210^{2}10410^{4}10610^{6}10−910^{-9}10−510^{-5}10−110^{-1}nnDouble word (bfloat16)10210^{2}10410^{4}10610^{6}10−910^{-9}10−510^{-5}10−110^{-1}nnTriple word (bfloat16)
MI100MI250MI300XB200
Figure 6. Multi-word arithmetic experiment presented by Mary and Mikaitis (Mary and Mikaitis, 2025, Sec. 5) on the simulation of various tensor cores. Relative norm-wise errors of matrix multiplication, compared with a default MATLAB binary64 multiplication, are shown. The input matrices to the GEMM are A∈R10×nA\in R^{10\times n} and B∈Rn×10B\in R^{n\times 10}. These matrices are multiplied with a multi-word algorithm (Mary and Mikaitis, 2025, Sec. 4) by splitting them into several bffloat16 words.

8.2. Broadband Signal Processing Algorithms

Many signal processing algorithms rely heavily on GEMMs, including applications in tensor decomposition, polynomial matrix algebra, and multiple-input multiple-output (MIMO) systems. Here, we investigate the impact of reduced-precision GEMM implementations on the output of a sequential matrix diagonalisation (SMD) (Redif et al., 2015) algorithm, a polynomial eigenvalue decomposition (PEVD) method used in broadband array processing applications (d’Olne et al., 2025; Neo et al., 2023).

We consider a 𝑹⁡(z)=∑τ𝐑⁡[τ]​z−τ∈ℝ8×8{{\boldsymbol{R}}(z)}=\sum_{\tau}{\mathbf{R}}[\tau]z^{-\tau}\in\mathbb{R}^{8\times 8}, where τ∈ℤ\tau\in\mathbb{Z}, with polynomial order 88 as an example case. The off-diagonal energy threshold ϵ\epsilon and the truncation parameter μ\mu are both set 10−410^{-4} with algorithm allowed to iterate for 250250 iterations. The reference result is obtained from a MATLAB double-precision implementation of the complete SMD algorithm. For the GEMM-based implementations, only the computationally dominant GEMM operation, i.e., updating the paraunitary and partially diagonalized matrices, is replaced by bit-accurate tensor and matrix core models representing different GPU architectures.

In this demonstration, in addition to CDNA 1, 2, and 3 architectures, we also consider the B200 NVIDIA GPU for both fp16 and bf16 input formats. The dominant eigenvalue and the condition number of the example polynomial matrix are shown in Figure 7. Although no visible difference can be observed between the eigenvalue estimates obtained using different GEMM models, the difference is non-zero, as illustrated in Figure 8 and  9. The frequency-dependent variations i.e., the dip around certain frequencies, are difficult to attribute to any particular feature because the error propagation depends on the repeated perturbation of the intermediate matrix 𝑫⁡(z){\boldsymbol{D}}(z) during subsequent SMD iterations. These perturbations may either accumulate towards or away from the reference solution at different frequencies, and therefore, require a separate analysis.

The absolute difference between the GEMM models, presented in Fig. 9, we can see that MI210 architecture exhibits comparatively larger differences from the other models, which may be attributed to its sequential fused multiply-add (SFMA) implementation and lack of subnormal number support. For fp16 inputs, the B200 shows the smallest difference with respect to MI100. This behaviour may be related to the two additional alignment bits available in B200, which provide behaviour closer to correctly rounded arithmetic compared with the other architectures.

For bf16 inputs, the MI300 and MI100 models exhibit almost identical behaviour. A possible explanation is that lower product significand precision reduces the difference between the finite and full-precision multi-term adders. This is further supported by the use of RNE rounding in both architectures. For example, bf16 products have lower significand precision, making them less susceptible to differences introduced by the 24-fractional-bit accumulator used by CDNA 3. In contrast, fp16 products have 22-bit significand precision and are therefore more susceptible to such differences. Since the observed difference is small but not exactly zero, this suggests the presence of rare out-of-range alignment cases.

It should be noted that these results are not conclusive. For a different polynomial matrix, the observed behaviour may differ because the iterative nature of the SMD algorithm causes numerical perturbations to be repeatedly propagated through subsequent iterations. These perturbations may lead the algorithm either closer to or further from the reference output at each frequency. Nevertheless, this example demonstrates that small numerical differences introduced by reduced-precision GEMM implementations can influence the final result of iterative matrix algorithms. In subspace-based applications (Khattak et al., 2025), such deviations may manifest as subspace leakage, where errors in the estimated eigenvectors or invariant subspaces degrade the separation between signal and noise components, thereby deteriorating the performance. In NN-based polynomial eigenvalue extraction (Hassan et al., 2024), these differences may be more pronounced if both the SMD algorithm and NN training and inference are performed on the matrix multiplier across different GPU architectures. Experiments with fp8 precision were also conducted; however, the estimated dominant eigenvalues exhibited significantly larger deviations from the double-precision reference. Therefore, fp8 precision is considered unsuitable for the considered example case.

This example demonstrates that seemingly minor numerical differences in low precision GEMM implementations can propagate through iterative matrix algorithms and influence the final eigenvalue, and therefore the corresponding eigenvector, estimates. Therefore, numerical characteristics of tensor and matrix core architectures must be carefully considered when deploying reduced-precision GEMM operations in signal processing applications.

Refer to caption
Figure 7. Dominant eigenvalue of the example parahermitian polynomial matrix estimated using the SMD algorithm. The MATLAB double-precision GEMM implementation is used as the reference, while the remaining curves correspond to GEMM computations performed on different GPU matrix and tensor core architectures.
Refer to caption
Figure 8. Absolute difference in the estimated dominant eigenvalue of the example parahermitian matrix between the reference model (in double precision), and the various matrix and tensor-core models in fp16 and bf16 input formats.
Refer to caption
Figure 9. Magnitude of the pairwise difference in the dominant eigenvalue estimated using the SMD algorithm due to GEMM computations on different GPU tensor or matrix core models. Each curve represents |Δ​λi,j​(ej​Ω)|=|λi​(ej​Ω)−λj​(ej​Ω)||\Delta\lambda_{i,j}(\mathrm{e}^{\mathrm{j}\Omega})|=|\lambda_{i}(\mathrm{e}^{\mathrm{j}\Omega})-\lambda_{j}(\mathrm{e}^{\mathrm{j}\Omega})|, where ii and jj denote the GPU pair identified in the legend.

9. Conclusion

Matrix multipliers in AMD CDNA 1, CDNA 2, and CDNA 3 architectures support a wide range of low-precision input formats; however, their internal numerical behaviors contain diverse architectural features that are not publicly documented. While prior work has characterized these features (Xie et al., 2025), we systematically re-determine all features of CDNA 1, 2, and 3. Our results match those reported in the GitHub repository but differ from those reported in (Xie et al., 2025). We also provide the test vectors used to identify the numerical features, which can be used to characterize future architectures. Additionally, we provide flexible software models of the CDNA 1, 2, and 3 matrix cores that allow users to modify individual numerical features for numerical analysis. The models are validated in MATLAB against GPU results using two sampling strategies, with randomized test suites exceeding 10710^{7} samples for all supported input formats and associated probabilistic upper bounds on the mismatch probability. The models also work in GNU Octave and are accessible within Python-based workflows through available language interfaces.

Our testing analysis reveals that the numerical features of AMD CDNA matrix cores vary significantly across generations and differ fundamentally from previously reported behaviors of NVIDIA GPU tensor cores. Specifically, CDNA 1 employs full-precision accumulation, CDNA 2 implements sequential FMA-based pair-wise accumulation approach, while CDNA 3 introduces a distinct numerical behavior that differs from both earlier AMD architectures and NVIDIA implementations.

Finally, we demonstrate the practical impact of these numerical differences through two application scenarios: multi-word arithmetic and polynomial matrix algebra. In both cases, small variations in matrix multiplication behavior can accumulate over repeated operations and lead to different computational outcomes. These examples emphasize the importance of understanding the numerical characteristics of hardware accelerators when developing applications.

10. Acknowledgment

We thank The COSmology MAchine (COSMA) support team at Durham University for providing the access to MI100/210/300A/300X. We thank Argonne National Laboratory for providing access to MI100/250/300 GPUs. FK and MM are funded by the EPSRC grant “Informing Future Numerical Standards by Determining Features of Non-Standard Mathematical Hardware”, ref. UKRI151. CG was supported by the Office of Advanced Scientific Computing Research, Office of Science, U.S. Department of Energy, under Contract DE-AC02-06CH11357.

References

  • A. Khattak and Mikaitis (2025) F. A. Khattak and M. Mikaitis Generalized methodology for determining numerical features of hardware floating-point matrix multipliers: part I. In 2025 IEEE High Performance Extreme Computing Conference (HPEC), Wakefield, MA, USA. External Links: Document Cited by: §1, §3, §3, item 1, §4.1.1, §4.1.2, §4.
  • Alexandridis and Dimitrakopoulos (2025) K. Alexandridis and G. Dimitrakopoulos Online alignment and addition in multiterm floating-point adders. IEEE Transactions on Very Large Scale Integration (VLSI) Systems 33 (4), pp. 1182–1186. External Links: Document Cited by: §2.
  • d’Olne et al. (2025) E. d’Olne, V. W. Neo, and P. A. Naylor PEVD-based speech enhancement in wireless acoustic sensor networks. IEEE Transactions on Audio, Speech and Language Processing 33 (), pp. 1467–1478. External Links: Document Cited by: §8.2.
  • Fasi et al. (2021) M. Fasi, N. J. Higham, M. Mikaitis, and S. Pranesh Numerical behavior of NVIDIA tensor cores. PeerJ Computer Science 7, pp. e330. External Links: Document Cited by: §1, §3, §4.1.1.
  • Fasi and Mikaitis (2023) M. Fasi and M. Mikaitis CPFloat: a C library for simulating low-precision arithmetic. ACM Trans. Math. Softw. 49 (2), pp. 18:1–18:32. External Links: ISSN 0098-3500, Document Cited by: §7.
  • Hassan et al. (2024) D. A. Hassan, Y. Egi, and S. Redif Neural networks for computing eigenvalues of parahermitian matrices. In 2024 32nd European Signal Processing Conference (EUSIPCO), Vol. , pp. 1307–1311. External Links: Document Cited by: §8.2.
  • Hickmann and Bradford (2019) B. Hickmann and D. Bradford Experimental analysis of matrix multiplication functional units. In 2019 IEEE 26th Symposium on Computer Arithmetic (ARITH), pp. 116–119. External Links: Document Cited by: §1, §3, item 1.
  • IEEE Computer Society (2019) IEEE standard for floating-point arithmetic, IEEE Std 754-2019 (revision of IEEE std 754-2008). Institute of Electrical and Electronics Engineers, Piscataway, NJ, USA. External Links: Document, ISBN 978-0-7381-5752-8 Cited by: §2.
  • Khattak et al. (2025) F. A. Khattak, M. Bakhit, I. K. Proudler, and S. Weiss Improved polynomial music algorithm for low-complexity and high-accuracy broadband angle of arrival estimation. In 2025 IEEE Statistical Signal Processing Workshop (SSP), Vol. , pp. 261–265. External Links: Document Cited by: §8.2.
  • Khattak and Mikaitis (2026) F. Khattak and M. Mikaitis Accurate models of nvidia tensor cores. ACM Trans. Archit. Code Optim. 23 (3). External Links: ISSN 1544-3566, Link, Document Cited by: item 2, §1, §2, §3, §3, item 1, §4.1.2, §4.1.3, §4.3.1, §4.3.2, §4.3.2, Table 3, §6.
  • Kulisch (2008) U. Kulisch Computer arithmetic and validity: theory, implementation, and applications. De Gruyter Studies in Mathematics, Vol. 33, Walter de Gruyter, Berlin, Germany. External Links: ISBN 9783110203189 Cited by: item 1.
  • Li et al. (2024) X. Li, A. Li, B. Fang, K. Swirydowicz, I. Laguna, and G. Gopalakrishnan FTTN: feature-targeted testing for numerical properties of NVIDIA & AMD matrix accelerators. In 2024 IEEE 24th International Symposium on Cluster, Cloud and Internet Computing (CCGrid), pp. 39–46. External Links: Document Cited by: §1, §3, §4.1.1, §4.1.2.
  • Mary and Mikaitis (2025) T. Mary and M. Mikaitis Error analysis of matrix multiplication with narrow range floating-point arithmetic. SIAM J. Sci. Comput. 47 (4), pp. B785–B800. External Links: Document Cited by: item 3, Figure 4, Figure 5, Figure 6, §8.1.
  • Mikaitis (2024) M. Mikaitis Monotonicity of multi-term floating-point adders. IEEE Trans. Comput. 73 (6), pp. 1531–1543. External Links: Document Cited by: §2, §4.1.2.
  • Neo et al. (2023) V. W. Neo, S. Redif, J. G. McWhirter, J. Pestana, I. K. Proudler, S. Weiss, and P. A. Naylor Polynomial eigenvalue decomposition for multichannel broadband signal processing: a mathematical technique offering new insights and solutions. IEEE Signal Processing Magazine 40 (7), pp. 18–37. External Links: Document Cited by: §8.2.
  • Redif et al. (2015) S. Redif, S. Weiss, and J. G. McWhirter Sequential matrix diagonalization algorithms for polynomial evd of parahermitian matrices. IEEE Transactions on Signal Processing 63 (1), pp. 81–89. External Links: Document Cited by: item 3, §8.2.
  • Tenca (2009) A. F. Tenca Multi-operand floating-point addition. In 2009 19th IEEE Symposium on Computer Arithmetic, pp. 161–168. External Links: Document Cited by: §2, item 1.
  • Valpey et al. (2025) B. Valpey, X. Li, S. Pai, and G. Gopalakrishnan An SMT formalization of mixed-precision matrix multiplication. In NASA Formal Methods, Cham, pp. 360–379. Cited by: §1.
  • Xie et al. (2025) P. Xie, Y. Wang, F. Yang, and M. Yang MMA-sim: bit-accurate reference model of tensor cores and matrix cores. arXiv preprint arXiv:2511.10909. External Links: 2511.10909, Document, Link Cited by: item 1, §1, §2, §3, §4.2.1, §4.2, §4.3.1, §4.3.1, §4.3.1, §4.3.2, §4.3.4, §4.3, §6, §9.